#!/usr/bin/env python

# jplayer - A widget library representation.

# Copyright (C) 2006 The Jahshaka Project
# Released under the GPL.
# For more information, see http://www.jahshaka.org.

#
# Mandatory for openpluginlib initialisation
#

import platform
import sys
import os.path
import linecache

# This is required for the executable generated from py2exe
# to use the preinstalled openlibraries python bindings

sys.path.append(os.environ['PYTHONPATH'])

# Initialise the dynamic linking functionality - note that the ubuntu amd64
# distro lacks the dl module, hence fallback to the non-symbolic values

if platform.system( ) == "Linux":
	try:
		import dl
		sys.setdlopenflags( dl.RTLD_NOW | dl.RTLD_GLOBAL )
	except:
		sys.setdlopenflags( 257 )

#
# Import and initialise the openpluginlib
#

import openpluginlib
openpluginlib.init( "" )

#
# Import the widgets lib and create the application
#

import jahwidgets
app = jahwidgets.application( )

class call_method_observer( openpluginlib.observer ):
	def __init__( self, method ):
		openpluginlib.observer.__init__( self )
		self.method = method

	def updated( self, subject ):
		self.method( subject )

def create_property_and_attach( handle, propname, observer ):
	prop = openpluginlib.property( openpluginlib.createKeyFromString( propname ) )
	prop.setAlwaysNotify( True )
	jahwidgets.connect( propname, handle, prop )
	prop.attach( observer )
	return prop

#
# Shared state for QT Event delegates
#

class media_state:

	# Constructor
	def __init__( self, player ):
		self.player = player
		self.append = 0
		self.backdrop = None
		self.logo = None
		self.desktop = 1
		self.timer = jahwidgets.timer( )
		self.other = 0
		self.playlist = [ ]
		self.rows = 1
		self.cols = 1

	# Set the optional logo for an empty playlist
	def set_logo( self, file ):
		self.logo = jahwidgets.create_media( file )
		self.logo.selectable = 0

	# Set the optional backdrop for the desktop view
	def set_backdrop( self, file, x, y, w, h ):
		self.backdrop = jahwidgets.create_media( file )
		self.backdrop.visible = 0
		self.backdrop.selectable = 0
		self.backdrop.t = 1
		self.backdrop_x = x
		self.backdrop_y = y
		self.backdrop_w = w
		self.backdrop_h = h

	# Reset the player
	def reset( self, all = 1 ):
		self.player.clear( )
		if all:
			self.playlist = [ ]

	# Process a list of strings as media objects
	def process_media( self, list = 0 ):
		# Iterate through each argument
		if list != 0:
			for i in list:
				media = jahwidgets.create_media( i )
				self.playlist.append( media )

		# Reload everything
		self.player.clear( )
		if self.backdrop is not None:
			self.player.push_back( self.backdrop )
		for media in self.playlist:
			self.player.push_back( media )

		# Change the active item if need be
		if len( self.playlist ) and self.player.active not in self.playlist:
			self.player.active = self.playlist[ 0 ]
			self.length_prop.set( self.player.active.length )

		# Decide which thing to show
		if self.desktop and len( self.playlist ) > 1:
			self.desktop = 1
			self.show_desktop( )
		elif len( self.playlist ):
			self.desktop = 0
			self.backdrop.visible = 0
			self.show_active( )
		elif self.logo is not None:
			self.desktop = 1
			self.player.play( self.logo )
			self.player.speed = 0

	# Delete the media item
	def delete( self, media ):
		if media in self.playlist:
			if media is self.player.active:
				self.play_next( )
			self.playlist.remove( media )
			self.process_media( )

	# Sort the playlist (problems with python list sort, so using a bubble for now)
	def sort_playlist( self ):
		if self.desktop and len( self.playlist ) > 1:
			for i in range( len( self.playlist ) ):
				for j in range( len( self.playlist ) ):
					if self.compare( self.playlist[ i ], self.playlist[ j ] ) < 0:
						t = self.playlist[ i ]
						self.playlist[ i ] = self.playlist[ j ]
						self.playlist[ j ] = t
			self.process_media( )

	# Work in progress - reorganise list after move
	def compare( self, a, b ):
		rh = self.player.height / self.rows
		cw = self.player.width / self.cols
		ra = ( a.y * self.player.height ) / rh
		rb = ( b.y * self.player.height ) / rh
		ca = ( a.x * self.player.width ) / cw
		cb = ( b.x * self.player.width ) / cw
		return ( ra * self.cols + ca ) - ( rb * self.cols + cb )

	# Locate and play next in list
	def play_next( self ):
		if len( self.playlist ):
			current = self.player.active
			next = self.playlist[ 0 ]
			found = 0
			for i in self.playlist:
				if found:
					next = i
					break
				elif i is current:
					found = 1
			self.player.active = next
			self.length_prop.set( self.player.active.length )
			self.show_active( )
			self.highlight_active( )

	# Locate and play previous in list
	def play_prev( self ):
		if len( self.playlist ):
			current = self.player.active
			prev = self.playlist[ len( self.playlist ) - 1 ]
			found = 0
			for i in self.playlist:
				if i is current:
					break
				else:
					prev = i
			self.player.active = prev
			self.length_prop.set( self.player.active.length )
			self.show_active( )
			self.highlight_active( )

	# Switch between desktop and player mode
	def toggle_desktop( self ):
		self.desktop = not self.desktop
		if self.desktop:
			self.reveal_desktop( )
		else:
			self.reveal_active( )

	# Animation into player mode
	def reveal_active( self ):
		self.reveal = reveal_active_timer( self )
		self.timer.delegate( self.reveal )
		self.timer.start( 10, 0 )

	# Animation into desktop mode
	def reveal_desktop( self ):
		self.reveal = reveal_desktop_timer( self )
		self.timer.delegate( self.reveal )
		self.timer.start( 10, 0 )

	# Go to player mode
	def show_active( self ):
		active = self.player.active

		if not self.desktop and active:
			for media in self.player:
				if media is not active:
					media.visible = 0
				else:
					media.visible = 1

			active.x = 0
			active.y = 0
			active.w = 1
			active.h = 1
			self.highlight_active( )
			self.player.show_tips = False

	# Resize all loaded media as a desktop
	def show_desktop( self, process = 1 ):
		# Determine number of media clips
		count = self.player.size( )

		# Reduce the count by the backdrop
		if self.backdrop is not None:
			count = count - 1

		# Default for one video, play it full resolution
		rows = 1
		cols = 1

		# Determine the optimal number of rows and columns
		while count > rows * cols:
			cols = cols * 2
			rows = rows * 2

		while count <= ( rows - 1 ) * cols:
			rows = rows - 1
			if count < ( cols - 1 ) * rows:
				cols = cols - 1

		# Set up the boxes
		offset_x = 0
		offset_y = 0
		width = 1.0 / cols 
		height = 1.0 / rows 
		size = min( width, height )

		# Handle the backdrop if it's registered
		if self.backdrop is not None:
			if count > 1:
				self.backdrop.visible = 1
				offset_x = self.backdrop_x
				offset_y = self.backdrop_y
				width = width * self.backdrop_w
				height = height * self.backdrop_h
				size = size * self.backdrop_h * 0.9
			else:
				self.backdrop.visible = 0

		# Determine where we are
		row = 0
		col = 0

		for media in self.player:
			if media.selectable:
				media.visible = 0
				media.t = 0.5

		for media in self.player:
			if media.selectable:
				media.visible = 1
				media.x = offset_x + col * width
				media.y = offset_y + row * height
				media.w = media.h = size
				col = col + 1
				if col >= cols:
					row = row + 1
					col = 0
				if media.dirty and process:
					self.player.refresh( )
					app.processEvents( 0 )

		self.rows = rows
		self.cols = cols

		self.player.show_tips = True
		self.highlight_active( )

	# Ensure all media items except the current one are faded
	def highlight_active( self ):
		active = self.player.active
		for media in self.player:
			if media.selectable:
				if media is active:
					media.t = 1
				else:
					media.t = 0.5
		self.player.refresh( )

#
# Timer delegate for revealing the animating the fully active media into view
#

class reveal_active_timer( jahwidgets.timer_delegate ):
	def __init__( self, state ):
		jahwidgets.timer_delegate.__init__( self )
		self.state = state
		self.count = 10
		media = self.state.player.active
		self.x = media.x / self.count
		self.y = media.y / self.count
		self.w = ( 1.0 - media.w ) / self.count
		self.h = ( 1.0 - media.h ) / self.count
		self.speed = self.state.player.speed
		self.state.player.speed = 0

	def timer( self ):
		media = self.state.player.active
		media.x = media.x - self.x
		media.y = media.y - self.y
		media.w = media.w + self.w
		media.h = media.h + self.h
		self.count = self.count - 1
		if self.count == 0:
			self.state.show_active( )
			if media.length > 1:
				self.state.player.speed = self.speed
		self.state.player.refresh( )
		return self.count != 0

#
# Timer delegate for revealing the animating the desktop into view
#

class reveal_desktop_timer( jahwidgets.timer_delegate ):
	def __init__( self, state ):
		jahwidgets.timer_delegate.__init__( self )
		self.state = state
		self.count = 10
		self.state.show_desktop( 0 )
		media = self.state.player.active
		self.x = media.x / self.count
		self.y = media.y / self.count
		self.w = ( 1.0 - media.w ) / self.count
		self.h = ( 1.0 - media.h ) / self.count
		media.x = media.y = 0
		media.w = media.h = 1
		self.speed = self.state.player.speed
		self.state.player.speed = 0

	def timer( self ):
		media = self.state.player.active
		media.x = media.x + self.x
		media.y = media.y + self.y
		media.w = media.w - self.w
		media.h = media.h - self.h
		self.count = self.count - 1
		if self.count == 0:
			self.state.show_desktop( 0 )
			if media.length > 1:
				self.state.player.speed = self.speed
		self.state.player.refresh( )
		return self.count != 0

#
# Drag n Drop delegate
#

class drop_delegate( jahwidgets.drop_delegate ):
	def __init__( self, state ):
		jahwidgets.drop_delegate.__init__( self )
		self.state = state

	def accept( self ):
		return 1

	def operate( self ):
		player = self.state.player
		if not self.state.append:
			self.state.reset( )
		self.state.desktop = 1
		self.state.process_media( self )
		if self.state.player.active.length > 1:
			self.state.player.speed = 1

#
# Mouse delegate
#

class mouse_delegate( jahwidgets.mouse_delegate ):
	def __init__( self, state ):
		jahwidgets.mouse_delegate.__init__( self )
		self.state = state
		self.jog_x = 0
		self.moving = 0
		self.orig_x = 0.0
		self.orig_y = 0.0
		self.start_x = 0.0
		self.start_y = 0.0

	def mousePressEvent( self ):
		self.accept( )
		media = self.state.player.find( self.x, self.y )
		active = self.state.player.active

		if media:
			if media is not active:
				self.state.player.active = media
				self.state.length_prop.set( self.state.player.active.length )
				self.state.highlight_active( )

		if self.isDoubleClick:
			self.state.toggle_desktop( )
		elif self.isRightButton:
			self.jog_speed = self.state.player.speed
			self.jog_x = self.x
			self.state.player.speed = 0
		elif self.isMiddleButton:
			self.state.player.speed = not self.state.player.speed
		elif self.isLeftButton and media:
			self.moving = media
			self.orig_x = self.moving.x
			self.orig_y = self.moving.y
			self.start_x = self.x 
			self.start_y = self.y

	# TODO: Take physical aspect ratio of the image into account
	def accept_x( self, x, w ):
		return w > 1.0 and x <= 0.0 and x + w >= 1.0

	def accept_y( self, y, h ):
		return h > 1.0 and y <= 0.0 and y + h >= 1.0

	# Move the selected item - we're either in desktop mode or player.
	# When in desktop mode, all moves are valid
	# When in player mode, a move is equivalent to a pan - moves are only 
	# valid when the image is larger than the preview and should be bound
	# such that preview shows no more aspect ratio related dead space than 
	# that needed to fit the window 
	def mouseMoveEvent( self ):
		self.accept( )
		if self.jog_x > 0:
			self.state.player.speed = 2 * ( ( self.x - self.jog_x ) / 20 )
		elif self.moving:
			x = self.orig_x + float( self.x - self.start_x ) / self.state.player.width 
			y = self.orig_y + float( self.y - self.start_y ) / self.state.player.height
			if self.state.desktop or self.accept_x( x, self.moving.w ):
				self.moving.x = x
			if self.state.desktop or self.accept_y( y, self.moving.h ):
				self.moving.y = y
			self.state.player.refresh( )

	def mouseReleaseEvent( self ):
		self.accept( )
		if self.jog_x:
			self.state.player.speed = self.jog_speed
			self.jog_x = 0
		elif self.moving:
			self.moving = 0
			self.state.sort_playlist( )

#
# Keyboard delegate
#

class key_delegate( jahwidgets.key_delegate ):
	def __init__( self, state ):
		jahwidgets.key_delegate.__init__( self )
		self.state = state
		self.full = False

	def keyPressEvent( self ):
		self.accept( )
		key = self.key
		if key == 65:		# 'A'
			self.state.append = not self.state.append
		elif key == 67:		# 'C'
			self.state.reset( )
			self.state.process_media( )
		elif key == 70:		# 'F'
			self.full = not self.full
			self.state.player.toggle_full_screen( )
			self.state.top.show( not self.full )
			self.state.left.show( not self.full )
			self.state.right.show( not self.full )
			self.state.bottom.show( not self.full )
		elif key == 78:		# 'N'
			self.state.play_next( )
		elif key == 79:		# 'O'
			state = self.state
			while state.other != 0:
				state = state.other.state
			state.other = jplayer( )
			state.other.process_media( )
			state.other.player.speed = 0
		elif key == 80:		# 'P'
			self.state.play_prev( )
		elif key == 81:		# 'Q'
			self.state.player.active.rotate += 90
			self.state.player.refresh( )
		elif key == 87:		# 'W'
			self.state.player.active.rotate -= 90
			self.state.player.refresh( )
		elif key == 32: 	# ' '
			self.state.player.speed = not self.state.player.speed
		elif key == 4096:	# Escape
			self.state.player.close_parent_window( )
		elif key == 4100:	# Return
			self.state.toggle_desktop( )
		elif key == 4103:	# Delete
			self.state.delete( self.state.player.active )
		elif key == 4114:	# Cursor Left
			self.state.play_prev( )
		elif key == 4116:	# Cursor Right
			self.state.play_next( )
		else:
			print key

	def keyReleaseEvent( self ):
		self.accept( )

#
# Wheel delegate
#

class wheel_delegate( jahwidgets.wheel_delegate ):
	def __init__( self, state ):
		jahwidgets.wheel_delegate.__init__( self )
		self.state = state
		self.jog_x = 0

	def wheelEvent( self ):
		if not self.state.desktop:
			self.accept( )
			media = self.state.player.find( self.x, self.y )
			offset = self.delta / 12.0
			diff = offset * 0.002

			if not media:
				media = self.state.player.active

			if media:
				if self.delta > 0 or ( self.delta < 0 and media.w > 0.01 and media.h > 0.01 ):
					media.w += diff
					media.h += diff

					if media.w > 1.0:
						media.x -= diff / 2.0
						media.y -= diff / 2.0

						x = media.x - ( ( float( self.x ) / float( self.state.player.width ) ) - 0.5 ) / 20
						y = media.y - ( ( float( self.y ) / float( self.state.player.height ) ) - 0.5 ) / 20

						if x <= 0 and x + media.w >= 1.0: media.x = x
						elif x > media.x: media.x = 0
						else: media.x = ( - media.w + 1.0 )

						if y <= 0 and y + media.h >= 1.0: media.y = y
						elif y > media.y: media.y = 0
						else: media.y = ( - media.h + 1.0 )
					else:
						media.x = ( 1.0 - media.w ) / 2.0
						media.y = ( 1.0 - media.h ) / 2.0

					self.state.player.refresh( )

#
# mouse delegates for drag and resize
#
class drag_delegate( jahwidgets.mouse_delegate ):
	def __init__( self, widget ):
		jahwidgets.mouse_delegate.__init__( self )
		self.widget=widget
		self.pressed=False

	def mousePressEvent( self ):
		self.mouseDownLocation=(self.globalX, self.globalY)
		self.widgetStartLocation=(self.widget.x, self.widget.y)
		self.pressed=True
		self.accept( )

	def mouseMoveEvent( self ):
		if not self.pressed:
			return
		
		deltaX=self.globalX-self.mouseDownLocation[0]
		deltaY=self.globalY-self.mouseDownLocation[1]
		self.widget.move( self.widgetStartLocation[0]+deltaX, self.widgetStartLocation[1]+deltaY )
		self.accept( )

	def mouseReleaseEvent( self ):
		self.pressed=False
		self.accept( )

class resize_delegate( jahwidgets.mouse_delegate ):
	def __init__( self, widget ):
		jahwidgets.mouse_delegate.__init__( self )
		self.widget=widget
		self.pressed=False

	def mousePressEvent( self ):
		self.mouseDownLocation=(self.globalX, self.globalY)
		self.widgetStartSize=(self.widget.width, self.widget.height)
		self.pressed=True
		self.accept( )

	def mouseMoveEvent( self ):
		if not self.pressed:
			return
		
		deltaX=self.globalX-self.mouseDownLocation[0]
		deltaY=self.globalY-self.mouseDownLocation[1]
		self.widget.resize( self.widgetStartSize[0]+deltaX, self.widgetStartSize[1]+deltaY )
		self.accept( )

	def mouseReleaseEvent( self ):
		self.pressed=False
		self.accept( )

#
# Creates an instance of the jplayer
#

def get_absolute_path( filename ):
	return os.path.join( os.path.abspath( sys.path[ 0 ] ), filename ).replace( '\\', '/' )

class delta_observer( openpluginlib.observer ):
	def __init__( self, delta, prop, min = 0.0, max = 1.0 ):
		openpluginlib.observer.__init__( self )
		self.delta = delta
		self.prop = prop
		self.min = min
		self.max = max
		
	def updated( self, subject ):
		val = self.prop.valueAsDouble() + self.delta
		if val < self.min: val = self.min
		if val > self.max: val = self.max
		self.prop.set( val )

class speed_observer( openpluginlib.observer ):
	def __init__( self, player, speed ):
		openpluginlib.observer.__init__( self )
		self.player = player
		self.speed = speed
		
	def updated( self, subject ):
		self.player.speed = self.speed

class timed_update( jahwidgets.timer_delegate ):
	def __init__( self, handler ):
		jahwidgets.timer_delegate.__init__( self )
		self.handler = handler

	def timer( self ):
		self.handler.player.refresh( )
		return True

class scrubbing_observer( openpluginlib.observer ):
	def __init__( self, player, prop ):
		openpluginlib.observer.__init__( self )
		self.player = player
		self.prop = prop
		self.speed = player.speed
		self.timed_update = timed_update( self )
		self.timer = jahwidgets.timer( )

	def updated( self, subject ):
		if ( self.prop.valueAsBool() ):
			self.timer.delegate( self.timed_update )
			self.timer.start( 200, 0 )
			self.speed = self.player.speed
			self.player.speed = 0
		else:
			self.timer.stop( )
			self.player.refresh( )
			self.player.speed = self.speed

class mute_observer( openpluginlib.observer ):
	def __init__( self, mute_prop, player ):
		openpluginlib.observer.__init__( self )
		self.player = player
		self.mute_prop = mute_prop

	def updated( self, subject ):
		speed = self.player.speed
		self.player.mute = self.mute_prop.valueAsBool()
		if speed != 0: self.player.speed = speed

class frame_observer( openpluginlib.observer ):
	def __init__( self, player, prop, position, absolute ):
		openpluginlib.observer.__init__( self )
		self.player = player
		self.prop = prop
		self.position = position
		self.absolute = absolute
		
	def updated( self, subject ):
		if ( self.absolute ):
			self.player.speed = 1
			self.prop.set( self.position )
		else:
			self.player.speed = 0
			self.prop.set( self.position + self.prop.valueAsInt() )
		self.player.refresh( )

class to_end_observer( openpluginlib.observer ):
	def __init__( self, player, frame_prop, length_prop ):
		openpluginlib.observer.__init__( self )
		self.player = player
		self.frame_prop = frame_prop
		self.length_prop = length_prop

	def updated( self, subject ):
		self.frame_prop.set( self.length_prop.valueAsInt() )
		self.player.refresh( )

# more of an adapter than an observer; basically will allow propA to
# modify propB via a scale factor and vice-versa
class scaling_observer( openpluginlib.observer ):
	def __init__( self, propA, propA_accessor, convToA, propB, propB_accessor, convToB, scale ):
		openpluginlib.observer.__init__( self )
		self.propA = propA
		self.propA_accessor = propA_accessor
		self.convToA = convToA
		self.propB = propB
		self.propB_accessor = propB_accessor
		self.convToB = convToB
		self.scale = scale

		self.propA.attach( self )
		self.propB.attach( self )
		
	def updated( self, subject ):
		if self.propA == subject:
			val = self.scale * self.propA_accessor()
			# print "A changed: ", val
			self.propB.set( self.convToB( val ) )
		else:
			val = self.propB_accessor() / self.scale
			# print "B changed: ", val
			self.propA.set( self.convToA( val ) )

class set_always_on_top_observer( openpluginlib.observer ):
	def __init__( self, widget ):
		openpluginlib.observer.__init__( self )
		self.widget = widget
		self.state = False

	def updated( self, subject ):
		self.state = not self.state
		self.widget.set_always_on_top( self.state )

class jplayer:
	def __init__( self ):
		# Construct the window from the ui
		self.top_level = jahwidgets.create_borderless_toplevel()
		self.main_handle = jahwidgets.create_ui_from_file( get_absolute_path( "main.ui" ), self.top_level )
		self.toolbar_placeholder_handle = jahwidgets.find( "toolbar_placeholder", self.main_handle )
		self.toolbar_handle = jahwidgets.create_ui_from_file( get_absolute_path( "toolbar.ui" ), self.toolbar_placeholder_handle )
		self.player_handle = jahwidgets.find( "player", self.main_handle )
		self.top_level.show( True )
		self.top_level.resize( 600, 500 )

		# connect up drag and resize
		self.drag_delegate=drag_delegate( self.top_level )
		self.titlebar_left=jahwidgets.find( "titlebar_left", self.main_handle )
		self.titlebar_left.register_mouse_delegate( self.drag_delegate )
		self.titlebar_center=jahwidgets.find( "titlebar_center", self.main_handle )
		self.titlebar_center.register_mouse_delegate( self.drag_delegate )

		self.resize_delegate=resize_delegate( self.top_level )
		self.grab=jahwidgets.find( "grab", self.toolbar_handle )
		self.grab.register_mouse_delegate( self.resize_delegate )

		# Create the player
		self.player = jahwidgets.player( self.player_handle )
		self.player.loop = True

		# A common object for holding shared state between the delegates
		self.state = media_state( self.player )

		# Set the logo and backdrop
		self.state.set_logo( get_absolute_path( "jplayer-back.jpg" ) )
		self.state.set_backdrop( get_absolute_path( "jplayer-back.jpg" ), 0.05, 0.05, 0.9, 0.9 )

		# Register the drag n drop delegate
		self.dnd = drop_delegate( self.state )
		self.player.register_drop_delegate( self.dnd )

		# Register the mouse delegate
		self.mouse = mouse_delegate( self.state )
		self.player.register_mouse_delegate( self.mouse )

		# Register the wheel delegate
		self.wheel = wheel_delegate( self.state )
		self.player.register_wheel_delegate( self.wheel )

		# Register the keyboard delegate
		self.state.top = jahwidgets.find( "frame15", self.main_handle )
		self.state.left = jahwidgets.find( "frame16", self.main_handle )
		self.state.right = jahwidgets.find( "frame17", self.main_handle )
		self.state.bottom = self.toolbar_handle
		self.key = key_delegate( self.state )
		self.player.register_key_delegate( self.key )

		# setup some properties
		self.play_prop = openpluginlib.property( openpluginlib.createKeyFromString( "play" ) )
		self.play_prop.setAlwaysNotify( True )
		self.stop_prop = openpluginlib.property( openpluginlib.createKeyFromString( "stop" ) )
		self.stop_prop.setAlwaysNotify( True )
		self.ffwd_prop = openpluginlib.property( openpluginlib.createKeyFromString( "ffwd" ) )
		self.ffwd_prop.setAlwaysNotify( True )
		self.rew_prop = openpluginlib.property( openpluginlib.createKeyFromString( "rew" ) )
		self.rew_prop.setAlwaysNotify( True )
		self.start_prop = openpluginlib.property( openpluginlib.createKeyFromString( "start" ) )
		self.start_prop.setAlwaysNotify( True )
		self.end_prop = openpluginlib.property( openpluginlib.createKeyFromString( "end" ) )
		self.end_prop.setAlwaysNotify( True )
		self.frame_prop = openpluginlib.property( openpluginlib.createKeyFromString( "frame" ) )
		self.scrubbing_prop = openpluginlib.property( openpluginlib.createKeyFromString( "scrubbing" ) )
		self.scrubbing_prop.setAlwaysNotify( True )
		self.length_prop = openpluginlib.property( openpluginlib.createKeyFromString( "length" ) )
		self.state.length_prop = self.length_prop
		self.volume_prop = openpluginlib.property( openpluginlib.createKeyFromString( "volume" ) )
		self.unscaled_volume_prop = openpluginlib.property( openpluginlib.createKeyFromString( "unscaled_volume" ) )
		self.mute_prop = openpluginlib.property( openpluginlib.createKeyFromString( "mute" ) )

		# hook up the controls
		jahwidgets.connect( "frame:value", self.toolbar_handle, self.frame_prop )
		jahwidgets.connect( "frame:pressed", self.toolbar_handle, self.scrubbing_prop )
		self.scrubbing_prop.attach( scrubbing_observer( self.player, self.scrubbing_prop ) )
		jahwidgets.connect( "player:frame", self.player_handle, self.frame_prop )
		jahwidgets.connect( "frame:maxValue", self.toolbar_handle, self.length_prop )
		jahwidgets.connect( "player:length", self.player_handle, self.length_prop )

		jahwidgets.connect( "player:volume", self.player_handle, self.volume_prop )
		jahwidgets.connect( "volume:value", self.toolbar_handle, self.unscaled_volume_prop )
		self.so = scaling_observer( self.volume_prop, self.volume_prop.valueAsDouble, float,
					    self.unscaled_volume_prop, self.unscaled_volume_prop.valueAsInt, int,
					    100.0 )
		self.volume_prop.set( 1.0 )
		jahwidgets.connect( "help:on", self.toolbar_handle, self.mute_prop )
		self.mute_prop.attach( mute_observer( self.mute_prop, self.player ) )

		jahwidgets.connect( "play:clicked", self.toolbar_handle, self.play_prop )
		self.play_prop.attach( speed_observer( self.player, 1 ) )
		jahwidgets.connect( "stop:clicked", self.toolbar_handle, self.stop_prop )
		self.stop_prop.attach( speed_observer( self.player, 0 ) )

		jahwidgets.connect( "start:clicked", self.toolbar_handle, self.start_prop )
		self.start_prop.attach( frame_observer( self.player, self.frame_prop, 0, True ) )
		jahwidgets.connect( "end:clicked", self.toolbar_handle, self.end_prop )
		self.end_prop.attach( to_end_observer( self.player, self.frame_prop, self.length_prop ) )
		jahwidgets.connect( "ffwd:clicked", self.toolbar_handle, self.ffwd_prop )
		self.ffwd_prop.attach( frame_observer( self.player, self.frame_prop, 12, False ) )
		jahwidgets.connect( "rewind:clicked", self.toolbar_handle, self.rew_prop )
		self.rew_prop.attach( frame_observer( self.player, self.frame_prop, -12, False ) )

		# connect up the window buttons
		create_property_and_attach( self.main_handle, "close:clicked", call_method_observer( lambda a: app.quit() ) )
		create_property_and_attach( self.main_handle, "minimize:clicked", call_method_observer( lambda a: app.minimize() ) )
		create_property_and_attach( self.main_handle, "maximize:clicked", call_method_observer( lambda a: app.maximize() ) )

		# an always on top button
		create_property_and_attach( self.main_handle, "show_hide:clicked", set_always_on_top_observer( self.top_level ) )

	def process_media( self, list = 0 ):
		self.state.process_media( list )

#
# Main method
#
def traceit(frame, event, arg):
	if event == "line":
		lineno = frame.f_lineno
		filename = frame.f_globals["__file__"]
		if filename == "<stdin>":
			filename = "traceit.py"
		if filename.endswith(".pyc") or filename.endswith(".pyo"):
			filename = filename[:-1]
		name = frame.f_globals["__name__"]
		line = linecache.getline(filename, lineno)
		print "%s:%s: %s" % (name, lineno, line.rstrip())
	return traceit

def main( ):
	# Drop the first argument (script name) - remainder are video/audio sources
	sys.argv.pop( 0 )

	# setup some styling
	if not app.setStyle( "v3" ):
		print "failed to load motif theme"

	# Turn on trace if required
	if len( sys.argv ) >= 1 and sys.argv[ 0 ] == "--trace":
		sys.settrace( traceit )
		sys.argv.pop( 0 )

	# find the main jplayer instance
	player = jplayer( )

	# Process command line arguments
	player.process_media( sys.argv )

	# let the user know the name of the app
	app.setCaption( "jplayer" )

	# Execute the application
	app.start_event_loop( )
	
	openpluginlib.uninit( )

if __name__ == "__main__":
	main( )
